﻿package  
{
	import flash.utils.Dictionary;
	
	public class Bezier 
	{
		private static var memop:Dictionary = new Dictionary();
		private static var memod:Dictionary = new Dictionary();
		
		public static function getLineInterpolatedFunction(t:Number):Function {
			if (memop[t] != undefined) {
				return memop[t] as Function;
			}
			var _t:Number = 1.0 - t;
			var  t_pow2:Number =  t *  t;
			var _t_pow2:Number = _t * _t;
			var k0:Number =       t_pow2 *  t;       //t^3
			var k1:Number = 3.0 * t_pow2 * _t;       //3 * t^2 * (1 - t)
			var k2:Number = 3.0 * t      * _t_pow2;  //3 * t * (1 - t)^2 
			var k3:Number =      _t_pow2 * _t;       //(t - 1)^3
			var fn:Function = function(p0:Number, p1:Number, p2:Number, p3:Number):Number {
				return k0 * p3 + k1 * p2 + k2 * p1 + k3 * p0;
			};
			memop[t] = fn;
			return fn;
		}
		
		public static function getLineDifferentiationInterpolatedFunction(t:Number):Function {
			if (memod[t] != undefined) {
				return memod[t] as Function;
			}
			var _t:Number = 1.0 - t;
			var k0:Number = 3.0 *        t * t;       //t^2
			var k1:Number = 3.0 * 2.0 *  t * _t;       //2 * t * (1 - t)
			var k2:Number = 3.0 *       _t * _t;       //(t - 1)^2
			var fn:Function = function(p0:Number, p1:Number, p2:Number, p3:Number):Number {
				return k0 * (p3 - p2) + k1 * (p2 - p1) + k2 * (p1 - p0);
			};
			memod[t] = fn;
			return fn;
		}
		
		public static function getLineInterpolatorFunction(p0:Number, p1:Number, p2:Number, p3:Number):Function {
			return function(t:Number):Number {
				return getLineInterpolatedFunction(t).call(null, p0, p1, p2, p3);
			};
		}
		
		public static function getLineDifferentiationInterpolatorFunction(p0:Number, p1:Number, p2:Number, p3:Number):Function {
			return function(t:Number):Number {
				return getLineDifferentiationInterpolatedFunction(t).call(null, p0, p1, p2, p3);
			};
		}
		
		public static function interpolateLine(p0:Number, p1:Number, p2:Number, p3:Number, t:Number):Number {
			return getLineInterpolatedFunction(t).call(null, p0, p1, p2, p3);
		}
		
		public static function interpolateDifferentiationLine(p0:Number, p1:Number, p2:Number, p3:Number, t:Number):Number {
			return getLineDifferentiationInterpolatedFunction(t).call(null, p0, p1, p2, p3);
		}
		
		public static function getPlaneInterpolatedFunction(u:Number, v:Number):Function {
			return function(
				p00:Number, p01:Number, p02:Number, p03:Number,
				p10:Number, p11:Number, p12:Number, p13:Number,
				p20:Number, p21:Number, p22:Number, p23:Number,
				p30:Number, p31:Number, p32:Number, p33:Number):Number {
				var bezier_u:Function = getLineInterpolatedFunction(u);
				var bezier_v:Function = getLineInterpolatedFunction(v);
				return bezier_v(
					bezier_u(p00, p01, p02, p03),
					bezier_u(p10, p11, p12, p13),
					bezier_u(p20, p21, p22, p23),
					bezier_u(p30, p31, p32, p33)
				);
			}
		}
		
		public static function getPlaneDifferentiationInterpolatedFunctionU(u:Number, v:Number):Function {
			return function(
				p00:Number, p01:Number, p02:Number, p03:Number,
				p10:Number, p11:Number, p12:Number, p13:Number,
				p20:Number, p21:Number, p22:Number, p23:Number,
				p30:Number, p31:Number, p32:Number, p33:Number):Number {
				var bezier_du:Function = getLineDifferentiationInterpolatedFunction(u);
				var bezier_v:Function = getLineInterpolatedFunction(v);
				return bezier_du(
					bezier_v(p00, p10, p20, p30),
					bezier_v(p01, p11, p21, p31),
					bezier_v(p02, p12, p22, p32),
					bezier_v(p03, p13, p23, p33)
				);
			}
		}
		
		public static function getPlaneDifferentiationInterpolatedFunctionV(u:Number, v:Number):Function {
			return function(
				p00:Number, p01:Number, p02:Number, p03:Number,
				p10:Number, p11:Number, p12:Number, p13:Number,
				p20:Number, p21:Number, p22:Number, p23:Number,
				p30:Number, p31:Number, p32:Number, p33:Number):Number {
				var bezier_u:Function = getLineInterpolatedFunction(u);
				var bezier_dv:Function = getLineDifferentiationInterpolatedFunction(v);
				return bezier_dv(
					bezier_u(p00, p01, p02, p03),
					bezier_u(p10, p11, p12, p13),
					bezier_u(p20, p21, p22, p23),
					bezier_u(p30, p31, p32, p33)
				);
			}
		}
		
		public static function getPlaneInterpolatorFunction(
			p00:Number, p01:Number, p02:Number, p03:Number,
			p10:Number, p11:Number, p12:Number, p13:Number,
			p20:Number, p21:Number, p22:Number, p23:Number,
			p30:Number, p31:Number, p32:Number, p33:Number):Function {
			return function(u:Number, v:Number):Number {
				return getPlaneInterpolatedFunction(u, v).call(null,
					p00, p01, p02, p03,
					p10, p11, p12, p13,
					p20, p21, p22, p23,
					p30, p31, p32, p33
				);
			};
		}		
		
		public static function getPlaneDifferentiationInterpolatorFunctionU(
			p00:Number, p01:Number, p02:Number, p03:Number,
			p10:Number, p11:Number, p12:Number, p13:Number,
			p20:Number, p21:Number, p22:Number, p23:Number,
			p30:Number, p31:Number, p32:Number, p33:Number):Function {
			return function(u:Number, v:Number):Number {
				return getPlaneDifferentiationInterpolatedFunctionU(u, v).call(null,
					p00, p01, p02, p03,
					p10, p11, p12, p13,
					p20, p21, p22, p23,
					p30, p31, p32, p33
				);
			};
		}		
		
		public static function getPlaneDifferentiationInterpolatorFunctionV(
			p00:Number, p01:Number, p02:Number, p03:Number,
			p10:Number, p11:Number, p12:Number, p13:Number,
			p20:Number, p21:Number, p22:Number, p23:Number,
			p30:Number, p31:Number, p32:Number, p33:Number):Function {
			return function(u:Number, v:Number):Number {
				return getPlaneDifferentiationInterpolatedFunctionV(u, v).call(null,
					p00, p01, p02, p03,
					p10, p11, p12, p13,
					p20, p21, p22, p23,
					p30, p31, p32, p33
				);
			};
		}		
		
		public static function interpolatePlane(
			p00:Number, p01:Number, p02:Number, p03:Number,
			p10:Number, p11:Number, p12:Number, p13:Number,
			p20:Number, p21:Number, p22:Number, p23:Number,
			p30:Number, p31:Number, p32:Number, p33:Number, u:Number, v:Number):Number {
			return getPlaneInterpolatedFunction(u, v).call(null,
				p00, p01, p02, p03,
				p10, p11, p12, p13,
				p20, p21, p22, p23,
				p30, p31, p32, p33
			);
		}
		
		public static function interpolatePlaneDifferentiationU(
			p00:Number, p01:Number, p02:Number, p03:Number,
			p10:Number, p11:Number, p12:Number, p13:Number,
			p20:Number, p21:Number, p22:Number, p23:Number,
			p30:Number, p31:Number, p32:Number, p33:Number, u:Number, v:Number):Number {
			return getPlaneDifferentiationInterpolatedFunctionU(u, v).call(null,
				p00, p01, p02, p03,
				p10, p11, p12, p13,
				p20, p21, p22, p23,
				p30, p31, p32, p33
			);
		}
		
		public static function interpolatePlaneDifferentiationV(
			p00:Number, p01:Number, p02:Number, p03:Number,
			p10:Number, p11:Number, p12:Number, p13:Number,
			p20:Number, p21:Number, p22:Number, p23:Number,
			p30:Number, p31:Number, p32:Number, p33:Number, u:Number, v:Number):Number {
			return getPlaneDifferentiationInterpolatedFunctionV(u, v).call(null,
				p00, p01, p02, p03,
				p10, p11, p12, p13,
				p20, p21, p22, p23,
				p30, p31, p32, p33
			);
		}
	}
}