﻿package 
{
	import flash.display.*;
	import flash.events.*;
	import flash.geom.*;
	
	[SWF(backgroundColor="#0088ff", frameRate="24",width="400",height="400" )]
	public class Main extends Sprite 
	{
		private var viewport:Sprite;
		private var viewportColorLayer:Sprite;
		private var viewportShadeLayer:Sprite;
		private var viewportRefrectionLayer:Sprite;
		
		private var teapotGeometry:Object;
		
		private var checkerTexture:BitmapData;
		private var linerLightTexture:BitmapData;
		private var refrectionTexture:BitmapData;
		
		[Embed(source = 'reflection.jpg')] private var RefrectionTextureSource:Class;
		
		public function Main():void {
			if (stage) init();
			else addEventListener(Event.ADDED_TO_STAGE, init);
		}
		
		private function init(e:Event = null):void {
			removeEventListener(Event.ADDED_TO_STAGE, init);
			stage.quality = StageQuality.LOW;
			
			initViewport();
			
			createTeapotGeometry(4);
			trace("vertices:" + (teapotGeometry.vertices.length / 3)
				+ " faces:" + (teapotGeometry.indices.length / 3));
			createCheckerTexture(256, 256, 4, 4, 0xff0000, 0xffffff);
			createLinerLightTexture(0x333333, 0xffffff);
			createRefrectionTexture(RefrectionTextureSource);
			
			var lightDirection:Vector3D = new Vector3D(-1, 0.2, 1);
			lightDirection.normalize();
			
			var context:Object = {
				rotation:0.0,
				lightDirection:lightDirection,
				diffuse:  { level:0.4 },
				specular: { level:2.0, shininess:20 }
			};
			renderScene(context);
			
			addEventListener(Event.ENTER_FRAME, function(e:Event):void {
				renderScene(context);
				context.rotation += 1.0 * 30 / stage.frameRate;
			});
		}
		
		private function renderScene(context:Object):void {
			var i:int;
			
			// matrix operations (make up a scenegraph)
			var trans:Matrix3D = new Matrix3D();
			trans.appendTranslation(0, 0, -1.2);
			trans.appendRotation(90, new Vector3D(1, 0, 0));
			trans.appendRotation(context.rotation * 2.0, new Vector3D(0, 1, 0));
			trans.appendRotation(context.rotation * 1.0, new Vector3D(1, 0, 0));
			trans.appendRotation(context.rotation * 0.5, new Vector3D(0, 0, 1));
			trans.appendTranslation(0, 0, 12.0);
			
			// transform in 3d space
			var vertices3d:Vector.<Number> = new Vector.<Number>();
			vertices3d.length = teapotGeometry.vertices.length;
			trans.transformVectors(teapotGeometry.vertices, vertices3d);
			
			var normals3d:Vector.<Number> = new Vector.<Number>();
			normals3d.length = teapotGeometry.normals.length;
			exportRotationElement(trans).transformVectors(teapotGeometry.normals, normals3d);
			
			// 3d space to screeen
			var proj:PerspectiveProjection = new PerspectiveProjection();
			proj.fieldOfView = 45;
			
			var vertices2d:Vector.<Number> = new Vector.<Number>();
			vertices2d.length = vertices3d.length / 3 * 2;
			
			var mappings2d:Vector.<Number> = new Vector.<Number>();
			mappings2d.length = teapotGeometry.mappings.length / 2 * 3;
			
			Utils3D.projectVectors(proj.toMatrix3D(), vertices3d, vertices2d, mappings2d);
			
			// culling
			var indices:Vector.<int> = teapotGeometry.indices;
			var vertexInUse:Vector.<Boolean> = new Vector.<Boolean>;
			vertexInUse.length = vertices3d.length / 3;
			indices = culledIndices(vertices2d, indices, vertexInUse);
			
			// z-sorting
			indices = zSortedIndices(vertices3d, indices);
			
			// transport texture uv
			var uvtSize:int = 3; // or 2 to disable perspective correction
			if (uvtSize == 2) {
				mappings2d = new Vector.<Number>();
				mappings2d.length = vertices2d.length;
			}
			for (i = 0; i < teapotGeometry.mappings.length / 2; i++) {
				if(vertexInUse[i]){
					mappings2d[i * uvtSize    ] = teapotGeometry.mappings[i * 2];
					mappings2d[i * uvtSize + 1] = teapotGeometry.mappings[i * 2 + 1];
				}
			}
			
			// lighting
			var uvshade:Vector.<Number> = linerLightMapping(vertices3d, normals3d, vertexInUse, 
				context.lightDirection, context.diffuse.level, context.specular.level, context.specular.shininess);
			
			// refrection mapping
			var reflections:Vector.<Number> = new Vector.<Number>();
			reflections.length = vertices2d.length;
			for (i = 0; i < reflections.length / 2; i++) {
				if(vertexInUse[i]){
					reflections[i * 2    ] = normals3d[i * 3    ] * 0.5 + 0.5;
					reflections[i * 2 + 1] = normals3d[i * 3 + 1] * 0.5 + 0.5;
				}
			}
			
			// -- start rendering --
			
			// render to color layer
			viewportColorLayer.graphics.clear();
			viewportColorLayer.graphics.beginBitmapFill(checkerTexture, null, true, true);
			viewportColorLayer.graphics.drawTriangles(vertices2d, indices, mappings2d, TriangleCulling.NEGATIVE);
			viewportColorLayer.graphics.endFill();
			
			// render to shading layer
			viewportShadeLayer.graphics.clear();
			viewportShadeLayer.graphics.beginBitmapFill(linerLightTexture, null, false, false);
			viewportShadeLayer.graphics.drawTriangles(vertices2d, indices, uvshade, TriangleCulling.NEGATIVE);
			viewportShadeLayer.graphics.endFill();
			
			// render to reflection layer
			viewportRefrectionLayer.graphics.clear();
			viewportRefrectionLayer.graphics.beginBitmapFill(refrectionTexture, null, false, false);
			viewportRefrectionLayer.graphics.drawTriangles(vertices2d, indices, reflections, TriangleCulling.NEGATIVE);
			viewportRefrectionLayer.graphics.endFill();
		}
		
		private function initViewport():void {
			viewport = new Sprite();
			viewport.x = stage.stageWidth / 2;
			viewport.y = stage.stageHeight / 2;
			stage.addEventListener(Event.RESIZE, function(e:Event):void {
				viewport.x = stage.stageWidth / 2;
				viewport.y = stage.stageHeight / 2;
			});
			
			viewportColorLayer = new Sprite();
			viewport.addChild(viewportColorLayer);
			
			viewportShadeLayer = new Sprite();
			viewport.addChild(viewportShadeLayer);
			viewportShadeLayer.blendMode = BlendMode.MULTIPLY;
			
			viewportRefrectionLayer = new Sprite();
			viewport.addChild(viewportRefrectionLayer);
			viewportRefrectionLayer.blendMode = BlendMode.ADD;
			
			addChild(viewport);
		}
		
		private function createTeapotGeometry(steps:int):void {
			teapotGeometry = {
				vertices: new Vector.<Number>(),
				normals:  new Vector.<Number>(),
				mappings: new Vector.<Number>(),
				indices:  new Vector.<int>()
			};
			
			var patches:Array = TeapotPatch.evalPatch(steps);
			var idxBase:int = 0;
			for (var i:int = 0; i < patches.length; i++) {
				var p:Object = patches[i];
				for (var v:int = 0; v <= p.stepsV; v++) {
					for (var u:int = 0; u <= p.stepsU; u++) {
						var vert:Object = p.grid[(p.stepsU + 1) * v + u];
						teapotGeometry.vertices.push(vert.x, vert.y, vert.z);
						teapotGeometry.normals.push(vert.nx, vert.ny, vert.nz);
						teapotGeometry.mappings.push(vert.u, vert.v); // No perspective option
						if (u != 0 && v != 0) {
							var p00:int = idxBase - (p.stepsU + 1) - 1;
							var p10:int = idxBase - (p.stepsU + 1);
							var p01:int = idxBase - 1;
							var p11:int = idxBase;
							teapotGeometry.indices.push(p00, p01, p10);
							teapotGeometry.indices.push(p11, p10, p01);
						}
						idxBase++;
					}
				}
			}
		}
		
		private function createCheckerTexture(width:int, height:int, stepsU:int, stepsV:int, color1:uint, color2:uint):void {
			var checker:Sprite = new Sprite();
			var cellWidth:Number = Number(width) / stepsU;
			var cellHeight:Number = Number(height) / stepsV;
			for (var v:int = 0; v < stepsV; v++) {
				for (var u:int = 0; u < stepsU; u++){
					checker.graphics.beginFill((v + u) % 2 == 0 ? color1 : color2);
					checker.graphics.drawRect(
						cellWidth * u,
						cellHeight * v,
						cellWidth,
						cellHeight
					);
					checker.graphics.endFill();
				}
			}
			checkerTexture = new BitmapData(width, height);
			checkerTexture.draw(checker);
		}
		
		private function createLinerLightTexture(color1:uint, color2:uint):void {
			var linerGrad:Sprite = new Sprite();
			var m:Matrix = new Matrix();
			m.createGradientBox(256, 1);
			linerGrad.graphics.beginGradientFill(GradientType.LINEAR,
				[color1, color2],
				null,
				null, m);
			linerGrad.graphics.drawRect(0, 0, 256, 1);
			linerGrad.graphics.endFill();
			linerLightTexture= new BitmapData(256, 1, true, 0);
			linerLightTexture.draw(linerGrad);
		}
		
		private function createRefrectionTexture(source:Class):void {
			var bmp:Bitmap = Bitmap(new source());
			refrectionTexture = bmp.bitmapData;
		}
		
		private function exportRotationElement(m:Matrix3D):Matrix3D {
			var mcomp:Vector.<Vector3D> = m.decompose(Orientation3D.AXIS_ANGLE);
			var mrot:Matrix3D = new Matrix3D();
			mrot.appendRotation(mcomp[1].w * 180 / Math.PI, new Vector3D(
				mcomp[1].x,
				mcomp[1].y,
				mcomp[1].z));
			return mrot;
		}
		
		private function culledIndices(vertices2d:Vector.<Number>, indices:Vector.<int>, vertexInUse:Vector.<Boolean>):Vector.<int> {
			var numFaces:int = indices.length / 3;
			var filteredIndices:Vector.<int> = new Vector.<int>();
			for (var i:int = 0; i < vertexInUse.length; i++) {
				vertexInUse[i] = false;
			}
			for (i = 0; i < numFaces; i++) {
				var a0:int = indices[i * 3    ];
				var a1:int = indices[i * 3 + 1];
				var a2:int = indices[i * 3 + 2];
				
				var x0:Number = vertices2d[a0 * 2    ];
				var x1:Number = vertices2d[a1 * 2    ];
				var x2:Number = vertices2d[a2 * 2    ];
				
				var y0:Number = vertices2d[a0 * 2 + 1];
				var y1:Number = vertices2d[a1 * 2 + 1];
				var y2:Number = vertices2d[a2 * 2 + 1];
				
				// V1 = (x1-x0, y1-y0, 0)
				// V2 = (x2-x0, y2-y0, 0)
				// Veye = (0, 0, 1)
				// (V1 x V2) * Veye
				if ((x1 - x0) * (y2 - y0) - (y1 - y0) * (x2 - x0) >= 0.0) {
					filteredIndices.push(indices[i * 3    ]);
					filteredIndices.push(indices[i * 3 + 1]);
					filteredIndices.push(indices[i * 3 + 2]);
					vertexInUse[a0] = true;
					vertexInUse[a1] = true;
					vertexInUse[a2] = true;
				}
			}
			return filteredIndices;
		}
		
		private function linerLightMapping(vertices3d:Vector.<Number>, normals3d:Vector.<Number>, vertexInUse:Vector.<Boolean>,
			lightDir:Vector3D, diffuseLevel:Number, specularLevel:Number, specularShininess:Number):Vector.<Number>
		{	
			var uvshade:Vector.<Number> = new Vector.<Number>();
			uvshade.length = normals3d.length / 3 * 2;
			
			var normal:Vector3D = new Vector3D();
			var eyeToVert:Vector3D = new Vector3D();
			for (var i:int = 0; i < normals3d.length / 3; i++) {
				if(vertexInUse[i]){
					normal.x = normals3d[i * 3    ];
					normal.y = normals3d[i * 3 + 1];
					normal.z = normals3d[i * 3 + 2];
					var diffuse:Number = 0.0;
					var specular:Number = 0.0;
					if(diffuseLevel > 0.0){
						diffuse = Math.max(
							-lightDir.dotProduct(normal), 0
						) * diffuseLevel;
					}
					if (specularLevel > 0.0) {
						eyeToVert.x = vertices3d[i * 3    ];
						eyeToVert.y = vertices3d[i * 3 + 1];
						eyeToVert.z = vertices3d[i * 3 + 2];
						eyeToVert.normalize();
						eyeToVert.incrementBy(lightDir);
						eyeToVert.normalize();
						specular = Math.pow(Math.max(
							-eyeToVert.dotProduct(normal), 0
						), specularShininess) * specularLevel;
					}
					uvshade[i * 2    ] = diffuse + specular;
					uvshade[i * 2 + 1] = 0.5;
				}
			}
			
			return uvshade;
		}

		
		private function zSortedIndices(vertices3d:Vector.<Number>, indices:Vector.<int>):Vector.<int> {
			var numFaces:int = indices.length / 3;
			var facesPointer:Vector.<int> = new Vector.<int>();
			facesPointer.length = numFaces;
			var zs:Vector.<Number> = new Vector.<Number>();
			zs.length = numFaces;
			for (var i:int = 0; i < numFaces; i++) {
				facesPointer[i] = i;
				var a0:int = indices[i * 3    ];
				var a1:int = indices[i * 3 + 1];
				var a2:int = indices[i * 3 + 2];
				var z:Number = vertices3d[a0 * 3 + 2] + vertices3d[a1 * 3 + 2] + vertices3d[a2 * 3 + 2];
				zs[i] = z;
			}
			facesPointer.sort(function(a:int, b:int):Number {
				return zs[b] - zs[a];
			});
			var sortedIndices:Vector.<int> = new Vector.<int>();
			sortedIndices.length = indices.length;
			for (i = 0; i < numFaces; i++) {
				var fp:int = facesPointer[i];
				sortedIndices[i * 3    ] = indices[fp * 3    ];
				sortedIndices[i * 3 + 1] = indices[fp * 3 + 1];
				sortedIndices[i * 3 + 2] = indices[fp * 3 + 2];
			}
			return sortedIndices;
		}
	}
	
}