﻿/*
 * Original teapot code copyright follows:
 */

/*
 * (c) Copyright 1993, Silicon Graphics, Inc.
 *
 * ALL RIGHTS RESERVED
 *
 * Permission to use, copy, modify, and distribute this software
 * for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that
 * both the copyright notice and this permission notice appear in
 * supporting documentation, and that the name of Silicon
 * Graphics, Inc. not be used in advertising or publicity
 * pertaining to distribution of the software without specific,
 * written prior permission.
 *
 * THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU
 * "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR
 * OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  IN NO
 * EVENT SHALL SILICON GRAPHICS, INC.  BE LIABLE TO YOU OR ANYONE
 * ELSE FOR ANY DIRECT, SPECIAL, INCIDENTAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER,
 * INCLUDING WITHOUT LIMITATION, LOSS OF PROFIT, LOSS OF USE,
 * SAVINGS OR REVENUE, OR THE CLAIMS OF THIRD PARTIES, WHETHER OR
 * NOT SILICON GRAPHICS, INC.  HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH LOSS, HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE POSSESSION, USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * US Government Users Restricted Rights
 *
 * Use, duplication, or disclosure by the Government is subject to
 * restrictions set forth in FAR 52.227.19(c)(2) or subparagraph
 * (c)(1)(ii) of the Rights in Technical Data and Computer
 * Software clause at DFARS 252.227-7013 and/or in similar or
 * successor clauses in the FAR or the DOD or NASA FAR
 * Supplement.  Unpublished-- rights reserved under the copyright
 * laws of the United States.  Contractor/manufacturer is Silicon
 * Graphics, Inc., 2011 N.  Shoreline Blvd., Mountain View, CA
 * 94039-7311.
 *
 * OpenGL(TM) is a trademark of Silicon Graphics, Inc.
 */

package
{
	public class TeapotPatch
	{
		private static const cpdata:Array = [
			{x: 0.2,    y: 0.0,    z: 2.7    },			
            {x: 0.2,    y:-0.112,  z: 2.7    },
            {x: 0.112,  y:-0.2,    z: 2.7    },
            {x: 0.0,    y:-0.2,    z: 2.7    },
            {x: 1.3375, y: 0.0,    z: 2.53125},
            {x: 1.3375, y:-0.749,  z: 2.53125},
            {x: 0.749,  y:-1.3375, z: 2.53125},
            {x: 0.0,    y:-1.3375, z: 2.53125},
            {x: 1.4375, y: 0.0,    z: 2.53125},
            {x: 1.4375, y:-0.805,  z: 2.53125},
            {x: 0.805,  y:-1.4375, z: 2.53125},
            {x: 0.0,    y:-1.4375, z: 2.53125},
            {x: 1.5,    y: 0.0,    z: 2.4    },
            {x: 1.5,    y:-0.84,   z: 2.4    },
            {x: 0.84,   y:-1.5,    z: 2.4    },
            {x: 0.0,    y:-1.5,    z: 2.4    },
            {x: 1.75,   y: 0.0,    z: 1.875  },
            {x: 1.75,   y:-0.98,   z: 1.875  },
            {x: 0.98,   y:-1.75,   z: 1.875  },
            {x: 0.0,    y:-1.75,   z: 1.875  },
            {x: 2,      y: 0.0,    z: 1.35   },
            {x: 2,      y:-1.12,   z: 1.35   },
            {x: 1.12,   y:-2,      z: 1.35   },
            {x: 0.0,    y:-2,      z: 1.35   },
            {x: 2,      y: 0.0,    z: 0.9    },
            {x: 2,      y:-1.12,   z: 0.9    },
            {x: 1.12,   y:-2,      z: 0.9    },
            {x: 0.0,    y:-2,      z: 0.9    },
            {x:-2,      y: 0.0,    z: 0.9    },
            {x: 2,      y: 0.0,    z: 0.45   },
            {x: 2,      y:-1.12,   z: 0.45   },
            {x: 1.12,   y:-2,      z: 0.45   },
            {x: 0.0,    y:-2,      z: 0.45   },
            {x: 1.5,    y: 0.0,    z: 0.225  },
            {x: 1.5,    y:-0.84,   z: 0.225  },
            {x: 0.84,   y:-1.5,    z: 0.225  },
            {x: 0.0,    y:-1.5,    z: 0.225  },
            {x: 1.5,    y: 0.0,    z: 0.15   },
            {x: 1.5,    y:-0.84,   z: 0.15   },
            {x: 0.84,   y:-1.5,    z: 0.15   },
            {x: 0.0,    y:-1.5,    z: 0.15   },
            {x:-1.6,    y: 0.0,    z: 2.025  },
            {x:-1.6,    y:-0.3,    z: 2.025  },
            {x:-1.5,    y:-0.3,    z: 2.25   },
            {x:-1.5,    y: 0.0,    z: 2.25   },
            {x:-2.3,    y: 0.0,    z: 2.025  },
            {x:-2.3,    y:-0.3,    z: 2.025  },
            {x:-2.5,    y:-0.3,    z: 2.25   },
            {x:-2.5,    y: 0.0,    z: 2.25   },
            {x:-2.7,    y: 0.0,    z: 2.025  },
            {x:-2.7,    y:-0.3,    z: 2.025  },
            {x:-3,      y:-0.3,    z: 2.25   },
            {x:-3,      y: 0.0,    z: 2.25   },
            {x:-2.7,    y: 0.0,    z: 1.8    },
            {x:-2.7,    y:-0.3,    z: 1.8    },
            {x:-3,      y:-0.3,    z: 1.8    },
            {x:-3,      y: 0.0,    z: 1.8    },
            {x:-2.7,    y: 0.0,    z: 1.575  },
            {x:-2.7,    y:-0.3,    z: 1.575  },
            {x:-3,      y:-0.3,    z: 1.35   },
            {x:-3,      y: 0.0,    z: 1.35   },
            {x:-2.5,    y: 0.0,    z: 1.125  },
            {x:-2.5,    y:-0.3,    z: 1.125  },
            {x:-2.65,   y:-0.3,    z: 0.9375 },
            {x:-2.65,   y: 0.0,    z: 0.9375 },
            {x:-2,      y:-0.3,    z: 0.9    },
            {x:-1.9,    y:-0.3,    z: 0.6    },
            {x:-1.9,    y: 0.0,    z: 0.6    },
            {x: 1.7,    y: 0.0,    z: 1.425  },
            {x: 1.7,    y:-0.66,   z: 1.425  },
            {x: 1.7,    y:-0.66,   z: 0.6    },
            {x: 1.7,    y: 0.0,    z: 0.6    },
            {x: 2.6,    y: 0.0,    z: 1.425  },
            {x: 2.6,    y:-0.66,   z: 1.425  },
            {x: 3.1,    y:-0.66,   z: 0.825  },
            {x: 3.1,    y: 0.0,    z: 0.825  },
            {x: 2.3,    y: 0.0,    z: 2.1    },
            {x: 2.3,    y:-0.25,   z: 2.1    },
            {x: 2.4,    y:-0.25,   z: 2.025  },
            {x: 2.4,    y: 0.0,    z: 2.025  },
            {x: 2.7,    y: 0.0,    z: 2.4    },
            {x: 2.7,    y:-0.25,   z: 2.4    },
            {x: 3.3,    y:-0.25,   z: 2.4    },
            {x: 3.3,    y: 0.0,    z: 2.4    },
            {x: 2.8,    y: 0.0,    z: 2.475  },
            {x: 2.8,    y:-0.25,   z: 2.475  },
            {x: 3.525,  y:-0.25,   z: 2.49375},
            {x: 3.525,  y: 0.0,    z: 2.49375},
            {x: 2.9,    y: 0.0,    z: 2.475  },
            {x: 2.9,    y:-0.15,   z: 2.475  },
            {x: 3.45,   y:-0.15,   z: 2.5125 },
            {x: 3.45,   y: 0.0,    z: 2.5125 },
            {x: 2.8,    y: 0.0,    z: 2.4    },
            {x: 2.8,    y:-0.15,   z: 2.4    },
            {x: 3.2,    y:-0.15,   z: 2.4    },
            {x: 3.2,    y: 0.0,    z: 2.4    },
            {x: 0.0,    y: 0.0,    z: 3.15   },
            {x: 0.8,    y: 0.0,    z: 3.15   },
            {x: 0.8,    y:-0.45,   z: 3.15   },
            {x: 0.45,   y:-0.8,    z: 3.15   },
            {x: 0.0,    y:-0.8,    z: 3.15   },
            {x: 0.0,    y: 0.0,    z: 2.85   },
            {x: 1.4,    y: 0.0,    z: 2.4    },
            {x: 1.4,    y:-0.784,  z: 2.4    },
            {x: 0.784,  y:-1.4,    z: 2.4    },
            {x: 0.0,    y:-1.4,    z: 2.4    },
            {x: 0.4,    y: 0.0,    z: 2.55   },
            {x: 0.4,    y:-0.224,  z: 2.55   },
            {x: 0.224,  y:-0.4,    z: 2.55   },
            {x: 0.0,    y:-0.4,    z: 2.55   },
            {x: 1.3,    y: 0.0,    z: 2.55   },
            {x: 1.3,    y:-0.728,  z: 2.55   },
            {x: 0.728,  y:-1.3,    z: 2.55   },
            {x: 0.0,    y:-1.3,    z: 2.55   },
            {x: 1.3,    y: 0.0,    z: 2.4    },
            {x: 1.3,    y:-0.728,  z: 2.4    },
            {x: 0.728,  y:-1.3,    z: 2.4    },
            {x: 0.0,    y:-1.3,    z: 2.4    },
            {x: 0.0,    y: 0.0,    z: 0      },
            {x: 1.425,  y:-0.798,  z: 0      },
            {x: 1.5,    y: 0.0,    z: 0.075  },
            {x: 1.425,  y: 0.0,    z: 0      },
            {x: 0.798,  y:-1.425,  z: 0      },
            {x: 0.0,    y:-1.5,    z: 0.075  },
            {x: 0.0,    y:-1.425,  z: 0      },
            {x: 1.5,    y:-0.84,   z: 0.075  },
            {x: 0.84,   y:-1.5,    z: 0.075  }
        ];
		
		private static const patchdata:Array = [
			//rim
            {idx:[102, 103, 104, 105,   4,   5,   6,   7,
			        8,   9,  10,  11,  12,  13,  14,  15], mx:true, my:true},
            //body
            {idx:[ 12,  13,  14,  15,  16,  17,  18,  19,
			       20,  21,  22,  23,  24,  25,  26,  27], mx:true, my:true},
            {idx:[ 24,  25,  26,  27,  29,  30,  31,  32,
			       33,  34,  35,  36,  37,  38,  39,  40], mx:true, my:true},
            //lid
            {idx:[ 96,  96,  96,  96,  97,  98,  99, 100,
			      101, 101, 101, 101,   0,   1,   2,   3], mx:true, my:true},
            {idx:[  0,   1,   2,   3, 106, 107, 108, 109,
			      110, 111, 112, 113, 114, 115, 116, 117], mx:true, my:true},
            //bottom
            {idx:[118, 118, 118, 118, 124, 122, 119, 121,
			      123, 126, 125, 120,  40,  39,  38,  37], mx:true, my:true},
            //handle
            {idx:[ 41,  42,  43,  44,  45,  46,  47,  48,
			       49,  50,  51,  52,  53,  54,  55,  56], mx:false, my:true},
            {idx:[ 53,  54,  55,  56,  57,  58,  59,  60,
			       61,  62,  63,  64,  28,  65,  66,  67], mx:false, my:true},
            //spout
            {idx:[ 68,  69,  70,  71,  72,  73,  74,  75,
			       76,  77,  78,  79,  80,  81,  82,  83], mx:false, my:true},
            {idx:[ 80,  81,  82,  83,  84,  85,  86,  87,
			       88,  89,  90,  91,  92,  93,  94,  95], mx:false, my:true}
		];
		
		public function TeapotPatch() {
			
		}
		
		public static function evalPatch(steps:int):Array {
			var patches:Array = [];
			for (var i:int = 0; i < patchdata.length; i++) {
				var idx:Array = patchdata[i].idx as Array;
				patches.push(createPatchGrid(i, steps, steps, false, false));
				if (patchdata[i].my) {
					patches.push(createPatchGrid(i, steps, steps, false, true));
				}
				if(patchdata[i].mx) {
					patches.push(createPatchGrid(i, steps, steps, true, false));
					if (patchdata[i].my) {
						patches.push(createPatchGrid(i, steps, steps, true, true));
					}
				}
			}
			return patches;
		}
		
		private static function createPatchGrid(patchIdx:int, stepsU:int, stepsV:int, reverseX:Boolean, reverseY:Boolean):Object {
			var flipping:Number = (reverseX ? -1 : 1) * (reverseY ? -1 : 1);
			var u:int, v:int;
			var cps_x:Array = [];
			var cps_y:Array = [];
			var cps_z:Array = [];
			var idx:Array = []
			
			if (flipping == -1) {
				for (v = 0; v < 4; v++) {
					for (u = 0; u < 4; u++) {
						idx[v * 4 + u] = patchdata[patchIdx].idx[v * 4 + (3 - u)];
					}
				}
			}
			else {
				idx = patchdata[patchIdx].idx;
			}
			
			for (var i:int = 0; i < idx.length; i++) {
				var cp:Object = cpdata[idx[i]];
				cps_x.push(cp.x * (reverseX ? -1:1));
				cps_y.push(cp.y * (reverseY ? -1:1));
				cps_z.push(cp.z);
			}
			
			var interp_x:Function = Bezier.getPlaneInterpolatorFunction.apply(null, cps_x);
			var interp_y:Function = Bezier.getPlaneInterpolatorFunction.apply(null, cps_y);
			var interp_z:Function = Bezier.getPlaneInterpolatorFunction.apply(null, cps_z);
			
			var interp_u_dx:Function = Bezier.getPlaneDifferentiationInterpolatorFunctionU.apply(null, cps_x);
			var interp_u_dy:Function = Bezier.getPlaneDifferentiationInterpolatorFunctionU.apply(null, cps_y);
			var interp_u_dz:Function = Bezier.getPlaneDifferentiationInterpolatorFunctionU.apply(null, cps_z);
			
			var interp_v_dx:Function = Bezier.getPlaneDifferentiationInterpolatorFunctionV.apply(null, cps_x);
			var interp_v_dy:Function = Bezier.getPlaneDifferentiationInterpolatorFunctionV.apply(null, cps_y);
			var interp_v_dz:Function = Bezier.getPlaneDifferentiationInterpolatorFunctionV.apply(null, cps_z);
			
			var positionOn:Function = function(u:Number, v:Number):Object {
				return {
					x:interp_x(uu, vv),
					y:interp_y(uu, vv),
					z:interp_z(uu, vv)
				};
			};
			var normalOn:Function = function(u:Number, v:Number):Object {
				return normalizedCrossProduct(
					{
						x: interp_u_dx(u, v),
						y: interp_u_dy(u, v),
						z: interp_u_dz(u, v)
					},
					{
						x: interp_v_dx(u, v),
						y: interp_v_dy(u, v),
						z: interp_v_dz(u, v)
					}
				);
			};
			
			var grid:Array = [];
			
			for (v = 0; v <= stepsV; v++) {
				for (u = 0; u <= stepsU; u++) {
					var uu:Number = Number(u) / stepsU;
					var vv:Number = Number(v) / stepsV;
					var pos:Object = positionOn(uu, vv);
					var nrm:Object = normalOn(uu, vv);
					if (nrm.x == 0 && nrm.y == 0 && nrm.z == 0) {
						var fixu:Number = (uu - 0.5) * 0.99 + 0.5;
						var fixv:Number = (vv - 0.5) * 0.99 + 0.5;
						nrm = normalOn(fixu, fixv);
					}
					grid.push( {
						x:pos.x, y:pos.y, z:pos.z,
						nx:nrm.x, ny:nrm.y, nz:nrm.z,
						u:flipping == 1 ? uu : (1.0 - uu),
						v:flipping == 1 ? vv : (1.0 - vv)
					});
				}
			}
			return {stepsU:stepsU, stepsV:stepsV, grid:grid};
		}
		
		private static function normalizedCrossProduct(v0:Object, v1:Object):Object {
			var x:Number = v0.y * v1.z - v0.z * v1.y;
			var y:Number = v0.z * v1.x - v0.x * v1.z;
			var z:Number = v0.x * v1.y - v0.y * v1.x;
			var norm:Number = Math.sqrt(x * x + y * y + z * z);
			if (norm == 0) {
				return { x:0, y:0, z:0 };
			}
			return {
				x:x / norm,
				y:y / norm,
				z:z / norm
			};
		}
	}
	
}